//*****************************************************************************
//*File:		QUEUE.C
//*Purpose: Provides a 'C object oriented implementation of a circular queue
//*			The queue is meant to be accessed using 4 functions:
//*			QueueInit - Must be called to before any other queue function
//*			GetWord	- Gets a word from the queue and returns the queue status
//*							Returns QUEUE_EMPTY if there is no data in the queue
//*			PutWord	- Puts a word into the queue and returns the queue status
//*							Returns QUEUE_FULL if there is no room to store data into queue
//*			GetQueueStatus - Returns the current queue status
//*
//*         PeakQueue - Gets the word at the beginning of the queue -- leaves the content
//*                   of the queue intact --  Returns QUEUE_EMPTY if there is no data in the queue
//*         GetNumberOfWords - Gets the current word count of the queue
//*         GetCapacity - Get the number of words the queue can hold
//*****************************************************************************
#define _QUEUE_C
#include <malloc.h>
#include <stdlib.h>
#include "expa.h"
#include "queue.h"


//Instantiate the queue, declared as static because the structure is
//considered PRIVATE.  All queue accesses must be performed via queue access
//functions.
static QueueType Queue;
static unsigned long NumberOfWords;
static SIZE_OF_WORD_QUEUE;

//*****************************************************************
//Function:	QueueInit
//Purpose:	Initializes all queue parameters.  It must be called before
//				queue is used!!!!  May be used to reset the queue.
//Parameters:	None
//Returns:		None
//****************************************************************
BOOLEAN QueueInit(UINT32 queueSize)
{
	//Initialize the ptrs to point to the start of the queue
	SIZE_OF_WORD_QUEUE = queueSize + 1;
	Queue.wordArray = (WORD*) malloc((unsigned) (SIZE_OF_WORD_QUEUE) * sizeof(WORD));
        if( Queue.wordArray == NULL ){
		return my_false;
  	}	
	Queue.wordPutPtr = Queue.wordArray;
   Queue.wordGetPtr = Queue.wordArray;
   Queue.wordEndPtr = Queue.wordArray + SIZE_OF_WORD_QUEUE;

   //Initialize the queue's status to empty
   Queue.Status = QUEUE_EMPTY;
   NumberOfWords = 0;
   return my_true;
}

//*****************************************************************
//Function:	GetQueueStatus	
//Purpose:	To allow the caller to check the queue's status
//				This function does not modify status.  It simply returns
//				the status stored in the private queue.
//Parameters:	None
//Returns:		QueueStatus - the queue status
//****************************************************************
QueueStatusType GetQueueStatus( void )
{
	return Queue.Status;
}

//*****************************************************************
//Function:	PutWord
//Purpose:	If there is room in the queue, stores the word in the queue
//				If the queue is full, the function returns
//Parameters:	WORD wordInput - word to store
//Returns:		QueueStatus - the queue status; either QUEUE_IN_USE or QUEUE_FULL
//****************************************************************
QueueStatusType PutWord( WORD wordInput )
{
	//If the queue is full, simply return
   if( Queue.Status == QUEUE_FULL )
   {
		return QUEUE_FULL;
   }

   //Otherwise, store the word
    *(Queue.wordPutPtr) = wordInput;

   //Increment the index
   Queue.wordPutPtr++;

   //Check array bounds and adjust the index if necessary
   if( Queue.wordPutPtr == Queue.wordEndPtr )
   {
   	Queue.wordPutPtr = Queue.wordArray;
	}

   //Update the queue status (Either QUEUE_IN_USE or QUEUE_FULL)
	if( Queue.wordPutPtr == Queue.wordGetPtr )
   {
   	Queue.Status = QUEUE_FULL;
	}
   else
   {
   	Queue.Status = QUEUE_IN_USE;
   }

   NumberOfWords++;

   //Return the queue's status
   return Queue.Status;
}

//*****************************************************************
//Function:	GetWord
//Purpose:	If there is data in the queue, returns the next word
//				If the queue is empty, the function returns the QueueStatus
//Parameters:	WORD *wordOutput - ptr where word is returned
//Returns:		QueueStatus - the queue status; either QUEUE_IN_USE or
//										QUEUE_EMPTY
//****************************************************************
QueueStatusType GetWord( WORD *wordOutput )
{
	//If queue is empty, return QUEUE_EMPTY
	if( Queue.Status == QUEUE_EMPTY )
   {
   	return QUEUE_EMPTY;
	}

   //Otherwise get the word
    *wordOutput = *(Queue.wordGetPtr);

   //Increment the index
   Queue.wordGetPtr++;

   //Check array bounds and adjust the index if necessary
   if( Queue.wordGetPtr == Queue.wordEndPtr )
   {
   	Queue.wordGetPtr = Queue.wordArray;
	}

   //Update the queue status (Either QUEUE_IN_USE or QUEUE_EMPTY)
	if( Queue.wordGetPtr == Queue.wordPutPtr )
   {
   	Queue.Status = QUEUE_EMPTY;
	}
   else
   {
   	Queue.Status = QUEUE_IN_USE;
   }

   NumberOfWords--;

	//Return the Queue's status
   return Queue.Status;
}

//*****************************************************************
//Function:	PeakQueue
//Purpose:	If there is data in the queue, returns the next word
//              but leaves the queue intact
//				If the queue is empty, the function returns the QueueStatus
//Parameters:	WORD *wordOutput - ptr where word is returned
//Returns:		QueueStatus - the queue status; either QUEUE_IN_USE or
//										QUEUE_EMPTY
//****************************************************************
QueueStatusType PeakQueue( WORD *wordOutput )
{
	//If queue is empty, return QUEUE_EMPTY
	if( Queue.Status == QUEUE_EMPTY )
   	{
   	return QUEUE_EMPTY;
	}

   //Otherwise get the word
    *wordOutput = *(Queue.wordGetPtr);

	//Return the Queue's status
   return Queue.Status;
}

//*****************************************************************
//Function:	GetNumberOfWords
//Purpose:	Return the number of words in queue
//Parameters:	none
//Returns:	Number of words in Queue -- 0 if empty		
//****************************************************************
unsigned long GetNumberOfWords(void)
{
	if( Queue.Status == QUEUE_EMPTY )
	{
		return 0;
	}
	return NumberOfWords;
}

unsigned long GetCapacity(void)
{
	return SIZE_OF_WORD_QUEUE;
}


