
#define _READRXNFILE_C

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include "expa.h"
#include "readRxnFile.h"
#include "expaList.h"
#include "util.h"

/*finite state machine*/
static void (*stateTable[])() = {rxnName, met, rxnType, coeff, newline, xmet, extype, reverse};
static STATE_TYPE currentState; /*state for finite state machine*/
static UINT32 lineNumber;
static BOOLEAN reversible, firstTime, firstTime2, endDummy, exDummy;
static UINT32 numberOfMets;
static UINT32 numberOfRxns;
static UINT32 numberOfXfluxes;
static FILE* fp;

/*linked lists*/
static struct metElt** metHeader;
static struct metElt** metCurrent;
static struct rxnElt** rxnHeader;
static struct rxnElt** rxnCurrent;
static char** xFluxes;
static XFLXTYPE* xFluxTypes;
static UINT32 xSize = 0;

/***************************************************************************************************************************************
 *
 * 	Function:	readRxnFile()
 *
 * 	Description:	Function constructs the stoichiometric matrix from a reaction input file. The form of the matrix is in the form
 * 			given in Schilling et al., i.e., the metabolites (rows) and reactions (columns) are entered in the order as 
 * 			they appear in the reaction file and the columns corresponding to the exchange fluxes are put at the end of the
 * 			matrix.
 *
 * 	Parameters:	inputFile -- the name of a reaction file of type .expa or .cel
 * 			stoickMat -- pointer to the matrix which will contain the stoichiometric matrix after the file has been read
 * 			xPtr -- pointer to an array which will contain the values of the exchange fluxes (input = -1, free = 0, output = 1)
 * 			nrows -- address of a variable which will contain the number of rows of the stoichiometric matrix
 * 			ncols -- address of a variable which will contain the number of column of the stoichiometric matrix
 * 			nxflxs -- address of a variable which will contain the number of exchange fluxes of the stoichiometric matrix
 *
 * 	Date:		April 24, 2004
 *
 * 	Author:		Steve Bell
 *
 * **************************************************************************************************************************************/

void readRxnFile(char* inputFile, XTYPE*** stoickMat, XFLXTYPE** xPtr, UINT32* nrows, 
				 UINT32* ncols, UINT32* nxflxs)
{
	UINT32 xflx;

	/*linked lists holding metabolites and reactions*/
	if( ( metHeader = (struct metElt**) malloc((unsigned) sizeof(struct metElt*)) ) == NULL ){
		fprintf(stderr, "Cannot allocate pointer for met header \n");
		exit(4);
	}
	if( ( metCurrent = (struct metElt**) malloc((unsigned) sizeof(struct metElt*)) ) == NULL ){
		fprintf(stderr, "Cannot allocate pointer for met current \n");
		exit(4);
	}
	if( ( *metHeader = (struct metElt*) malloc((unsigned) sizeof(struct metElt)) ) == NULL ){
		fprintf(stderr, "Cannot allocate pointer for met header \n");
		exit(4);
	}
	if( ( rxnHeader = (struct rxnElt**) malloc((unsigned) sizeof(struct rxnElt*)) ) == NULL ){
		fprintf(stderr, "Cannot allocate pointer for rxn header \n");
		exit(4);
	}
	if( ( rxnCurrent = (struct rxnElt**) malloc((unsigned) sizeof(struct rxnElt*)) ) == NULL ){
		fprintf(stderr, "Cannot allocate pointer for rxn current \n");
		exit(4);
	}
	if( ( *rxnHeader = (struct rxnElt*) malloc((unsigned) sizeof(struct rxnElt)) ) == NULL ){
		fprintf(stderr, "Cannot allocate pointer for rxn header \n");
		exit(4);
	}

	if( (fp = fopen(inputFile,"r")) == NULL ){
		printf("Unable to open file %s \n", inputFile);
		exit(1);
	}
	initialize();
	while(endDummy != my_true){
		stateTable[currentState](); /*the parser is modeled as a finite state machine*/
	}
	fclose(fp);

	makeStoickMat(stoickMat);

	*xPtr = xflxvector(numberOfXfluxes);
	for( xflx = 1; xflx <= numberOfXfluxes; xflx++ ){
		(*xPtr)[xflx] = xFluxTypes[xflx-1]; 
	}
	*nrows = numberOfMets;
	*ncols = numberOfRxns + numberOfXfluxes;
	*nxflxs = numberOfXfluxes;
	finish();
}

/***************************************************************************************************************************************
 *
 * 	Function:	makeStoickMat()
 *
 * 	Description:	Function constructs the stoichiometric matrix from the linked lists containing the metabolites and the 
 * 			reactions.
 *
 * 	Parameters:	stoickMat -- pointer to the matrix which will contain the stoichiometric matrix	
 *
 * 	Date:		April 24, 2004
 *
 * 	Author:		Steve Bell
 *
 * **************************************************************************************************************************************/

void makeStoickMat(XTYPE*** stoickMat)
{
	UINT32 m, r;
	XTYPE coeff;
	char** mets;
	char** rxns;
	char* met;
	XTYPE** sMat = xtypeMatrix( numberOfMets, numberOfRxns + numberOfXfluxes );
	if( ( mets = (char**) malloc((unsigned) (numberOfMets + 1) * sizeof(char*)) ) == NULL ){
		fprintf(stderr, "Cannot allocate met vector \n");
		exit(4);
	}
	for( m = 1; m <= numberOfMets; m++ ){
		for( r = 1; r <= (numberOfRxns + numberOfXfluxes); r++ ){
			sMat[m][r] = 0;
		}
	}
	if( ( rxns = (char**) malloc((unsigned) (numberOfRxns + 1) * sizeof(char*)) ) == NULL ){
		fprintf(stderr, "Cannot allocate met vector \n");
		exit(4);
	}
	m = 0;
	metListZeroth(metHeader, metCurrent);
	while( metListAdvance(metCurrent) ){
		m++;
		mets[m] = (*metCurrent)->metName;
	}
	if( m != numberOfMets  ){
		fprintf(stderr, "Error in mets \n");
		exit(2);
	}
	r = 0;
	rxnListZeroth(rxnHeader, rxnCurrent);
	while( rxnListAdvance(rxnCurrent) ){
		r++;
		rxns[r] = (*rxnCurrent)->rxnName;		
		rListZeroth( (*rxnCurrent)->headerPtr, (*rxnCurrent)->currentPtr );
		while( rListAdvance( (*rxnCurrent)->currentPtr ) ){
			met = (*(*rxnCurrent)->currentPtr)->metName;
			coeff = (*(*rxnCurrent)->currentPtr)->coeff;
			for( m = 1; m <= numberOfMets; m++ ){
				if( strcmp( met, mets[m] ) == 0 ){
					sMat[m][r] = coeff;
				}
			}
		}
	}
	if( r != numberOfRxns ){
		fprintf(stderr, "Error in rxns \n");
	}
	for( r = 1; r <= numberOfXfluxes; r++ ){
		met = xFluxes[r-1];
		for( m = 1; m <= numberOfMets; m++ ){
			if( strcmp( met, mets[m] ) == 0 ){
				sMat[m][numberOfRxns + r] = -1;
				break;
			}
		}
	}
	*stoickMat = sMat;

	if( (stoickDummy == my_true) && (verbose == my_true) ){
		printf("Reactions (%d): \n", numberOfRxns);
		for( r = 1; r <= numberOfRxns; r++ ){
			printf("%d.  %s \n",r, rxns[r]);
		}
		printf("Metabolites (%d): \n", numberOfMets);
		for( r = 1; r <= numberOfMets; r++ ){
			printf("%d.  %s \n",r, mets[r]);
		}
		printf("Exchange Fluxes (%d):\n", numberOfXfluxes);
		for( r = 0; r < numberOfXfluxes; r++ ){
			printf("%d.  %s \n", r + 1, xFluxes[r]);
		}
		printf("\n\n");
	}
	
} /*makeStoickMat()*/

/***************************************************************************************************************************************
 *
 * 	Function:	initialize()
 *
 * 	Description:	Function initializes the linked lists
 *
 * 	Parameters:	none
 *
 * 	Date:		April 24, 2004
 *
 * 	Author:		Steve Bell
 *
 * **************************************************************************************************************************************/

void initialize(void)
{
	endDummy = my_false;
	exDummy = my_false;
	lineNumber = 0;
	numberOfRxns = 0;
	numberOfMets = 0;
	numberOfXfluxes = 0;
	reversible = my_false;
	firstTime = my_true;
	firstTime2 = my_true;
	currentState = NEWLINE;
	metListInit(metHeader, metCurrent);
	if( rxnListInit(rxnHeader, rxnCurrent) == my_false ){
		fprintf(stderr, "Cannot allocate memory for reaction list \n");
		exit(4);
	}
} /*initialize()*/

/***************************************************************************************************************************************
 *
 * 	Function:	rxnName()
 *
 * 	Description:	Function enters the reaction name for the current row of the input file into the linked list
 * 			corresponding to reactions
 *
 * 	Parameters:	none
 *
 * 	Date:		April 24, 2004
 *
 * 	Author:		Steve Bell
 *
 * **************************************************************************************************************************************/

void rxnName(void)
{
	INT32 scanReturn;
	char *rxn;
	rxn = (char*) malloc((unsigned) 512 * sizeof(char));
	if( rxn == NULL ){
		fprintf(stderr, "Cannot allocate pointer");
		exit(1);
	}
	scanReturn = fscanf(fp, "%s", rxn);
	if( (scanReturn == 0) || (scanReturn == EOF) ){
		fprintf(stderr, "Error in input file, line %d \n", lineNumber);
		exit(2);
	}
	if( rxnListInsert(rxn, rxnCurrent) == my_false ){
		fprintf(stderr, "Cannot allocate memory for rxn %s, \n", rxn);
		exit(4);
	}
	numberOfRxns++;
	currentState = RXN_TYPE;
} /*rxnName()*/

/***************************************************************************************************************************************
 *
 * 	Function:	reverse()
 *
 * 	Description:	Function enters the reversed reaction into the linked list if the reaction 
 * 			corresponding to the current row in the input file is reversible.
 *
 * 	Parameters:	none
 *
 * 	Date:		April 24, 2004
 *
 * 	Author:		Steve Bell
 *
 * **************************************************************************************************************************************/

void reverse(void)
{
	char* newRxn;
	struct rxnElt* tempCurrent;
	newRxn = (char*) malloc((unsigned) 512 * sizeof(char));
	if( newRxn == NULL ){
		fprintf(stderr, "Cannot allocate reaction pointer \n");
		exit(4);
	}
	strcpy(newRxn, "rev");
    tempCurrent = *rxnCurrent;
	strcat(newRxn, (*rxnCurrent)->rxnName);
	if( rxnListInsert(newRxn, rxnCurrent) == my_false ){
		fprintf(stderr,"Cannot allocate memory for rxn %s \n", newRxn);
		exit(4);
	}
	numberOfRxns++;
	rListZeroth( tempCurrent->headerPtr, tempCurrent->currentPtr );
	while( rListAdvance( tempCurrent->currentPtr ) ){
			if( rListInsert( (*(tempCurrent->currentPtr))->metName, -(*(tempCurrent->currentPtr))->coeff,
				(*rxnCurrent)->currentPtr ) == my_false ){
				fprintf(stderr, "Cannot allocate memory for met %s \n", 
					(*(tempCurrent->currentPtr))->metName);
				exit(4);
			}
	}
	currentState = NEWLINE;
} /*reverse()*/

/***************************************************************************************************************************************
 *
 * 	Function:	met()
 *
 * 	Description:	Function enters the next metabolite into the linked list from the current row in the input file
 *
 * 	Parameters:	none
 *
 * 	Date:		April 24, 2004
 *
 * 	Author:		Steve Bell
 *
 * **************************************************************************************************************************************/

void met(void)
{
	INT32 inChar, length, scanReturn;
	char* metName = (char*) malloc((unsigned) 512 * sizeof(char));
	if( metName == NULL ){
		fprintf(stderr, "Cannot allocate pointer");
		exit(1);
	}
	scanReturn = fscanf(fp, "%s", metName);
	if( (scanReturn == 0) || (scanReturn == EOF) ){
		fprintf(stderr, "Error in input file, line %d \n", lineNumber);
		exit(2);
	}
	length = strlen(metName);
	if( metName[length - 1] == ';' ){
		metName[length - 1] = '\0';
	}
	(*(*rxnCurrent)->currentPtr)->metName = metName;
	if( !metListFind(metName, metHeader, metCurrent) ){
		while( metListAdvance( metCurrent ) ){/*insert at end*/
		}
		if( metListInsert(metName, metCurrent) == my_false ){
			fprintf(stderr, "Cannot allocate memory for met %s \n", metName);
			exit(4);
		}
		numberOfMets++;
	}
	/*remove whitespace at end of line*/
	while( ( ( inChar = fgetc(fp) ) == ' ') || inChar == '\t' ){
	}
	switch(inChar){
		case EOF:
			endDummy = my_true;
			break;
		case '\n':
			if( reversible ){
				currentState = REVERSE;
			}
			else{
				currentState = NEWLINE;
			}
			break;
		case 13: /*carriage return*/
			if( inChar == '\n' ){
			}
			else{
				ungetc(inChar, fp);
			}
			if( reversible ){
				currentState = REVERSE;
			}
			else{
				currentState = NEWLINE;
			}
			break;
		default:
			ungetc(inChar, fp);
			currentState = COEFF;
	}
} /*met()*/

/***************************************************************************************************************************************
 *
 * 	Function:	rxnType()
 *
 * 	Description:	Function records whether the reaction corresponding to the current row in the input file is reversible or
 * 			not.
 *
 * 	Parameters:	none
 *
 * 	Date:		April 24, 2004
 *
 * 	Author:		Steve Bell
 *
 * **************************************************************************************************************************************/

void rxnType(void)
{
	INT32 scanReturn;
	char* type = (char*) malloc((unsigned) 2 * sizeof(char));
	if( type == NULL ){
		fprintf(stderr, "Cannot allocate pointer");
		exit(1);
	}
	scanReturn = fscanf(fp, "%s", type);
	if( (scanReturn == 0) || (scanReturn == EOF) ){
		fprintf(stderr, "Error in input file, line %d \n", lineNumber);
		exit(2);
	}
	if( strcmp(type, "R") == 0 ){
		reversible = my_true;
	}
	else if( strcmp(type,"I") == 0 ){
		reversible = my_false;
	}
	else{
		fprintf(stderr, "Error, line %d: reaction type must be 'I' of 'R'\n", lineNumber);
		exit(2);
	}	
	currentState = COEFF;
	free((void*) type);
} /*rxnType()*/

/***************************************************************************************************************************************
 *
 * 	Function:	coeff()
 *
 * 	Description:	Function enters the coefficient of the stoichiometric matrix corresponding to the metabolite of the
 * 			current reaction into the linked list
 *
 * 	Parameters:	none
 *
 * 	Date:		April 24, 2004
 *
 * 	Author:		Steve Bell
 *
 * **************************************************************************************************************************************/

void coeff(void)
{
	INT32 scanReturn;
	XTYPE coeffValue;
	scanReturn = fscanf(fp, "%ld", &coeffValue);
	if( (scanReturn == 0) || (scanReturn == EOF) ){
		fprintf(stderr,"Error in input file, line %d \n", lineNumber);
		exit(2);
	}
	if( rListInsert((char*) NULL, coeffValue, (*rxnCurrent)->currentPtr) == my_false ){
		fprintf(stderr, "Cannot allocate memory for coefficient %d \n", coeffValue);
		exit(4);
	}
	currentState = MET;
} /*coeff()*/

/***************************************************************************************************************************************
 *
 * 	Function:	newline()
 *
 * 	Description:	Function performs utility chores when a newline character is encountered in the input file
 *
 * 	Parameters:	none
 *
 * 	Date:		April 24, 2004
 *
 * 	Author:		Steve Bell
 *
 * **************************************************************************************************************************************/

void newline(void)
{	
	int inChar;
	INT32 scanReturn;
	char* inString;
	inString = (char*) malloc((unsigned) 50 * sizeof(char));
	lineNumber++;
	if( inString == NULL ){
		fprintf(stderr, "Cannot allocate pointer");
		exit(1);
	}
	while( ( (inChar = getc(fp)) == '\n' ) || (inChar == 9) || (inChar == 13) ){
		lineNumber++;
	}
	ungetc(inChar, fp);
	if( (inChar = getc(fp)) == EOF ){
		endDummy = my_true;
	}
	else{
		if( firstTime2 == my_true ){
			if( ( inChar != '(' ) && ( inChar != '*' ) && ( inChar != '/' ) ){
				fprintf(stderr, "Error in input file, line %d \n", lineNumber);
				exit(1);
			}
		}
		if( inChar == '(' ){
			if( (inChar = getc(fp)) == '(' ){
				scanReturn = fscanf(fp, "%s", inString);
				if( (scanReturn == 0) || (scanReturn == EOF) ){
					fprintf(stderr, "Error in input file, line %d\n", lineNumber);
					exit(2);
				}
				inString = lowerCase(inString);
				if( ( strcmp(inString, "exchange") == 0 ) ||
					( strcmp(inString, "external") == 0 ) ){
					while( (inChar = getc(fp)) != '\n'){
					}
					exDummy = my_true;
					currentState = NEWLINE;
				}
				else if ( strcmp(inString, "internal") == 0 ){
					while( (inChar = getc(fp)) != '\n'){
					}
					firstTime2 = my_false;
					currentState = NEWLINE;
				}
				else{
					fprintf(stderr, "Error in input file line %d \n", lineNumber);
					exit(2);
				}
			}
			else{
				ungetc(inChar, fp);
				scanReturn = fscanf(fp, "%s", inString);
				if( (scanReturn == 0) || (scanReturn == EOF) ){
					fprintf(stderr, "Error in input file, line %d\n", lineNumber);
					exit(2);
				}
				inString = lowerCase(inString);
				if( ( strcmp(inString, "exchange") == 0 ) ||
					( strcmp(inString, "external") == 0 ) ){
					while( (inChar = getc(fp)) != '\n'){
					}
					exDummy = my_true;
					currentState = NEWLINE;
				}
				else if ( strcmp(inString, "internal") == 0 ){
					while( (inChar = getc(fp)) != '\n'){
					}
					firstTime2 = my_false;
					currentState = NEWLINE;
				}
				else{
					fprintf(stderr, "Error in input file line %d \n", lineNumber);
					exit(2);
				}
			}
		}
		else{ /*inChar != '('*/		
			if( (inChar == '*') || (inChar == '/') ){
				while(1){
					if( (inChar == '/') ){
						inChar = getc(fp);
						if( inChar == '/' ){
							while( ((inChar = getc(fp)) != '\n') && (inChar != EOF) ){
							}
							if( inChar == EOF ){
								ungetc(inChar, fp);
							}
						}
						else{
							fprintf(stderr, "Error in input file, line %d \n", lineNumber);
							exit(1);
						}
					}
					else{
						while( (inChar = getc(fp)) != '\n' ){
						}
					}
					inChar = getc(fp);
					if( (inChar != '*') && (inChar != '/') ){
						ungetc(inChar, fp);
						break;
					}
					lineNumber++;
				}
				currentState = NEWLINE;
			}
			else{
				ungetc(inChar,fp);
				if( exDummy == my_false ){
					currentState = RXN_NAME;
				}
				else{
					currentState = XMET;
				}
			}
		}
	}
	free((void*) inString);	
} /*newline()*/

/***************************************************************************************************************************************
 *
 * 	Function:	xmet()
 *
 * 	Description:	Function enters the exchange flux into the array 
 *
 * 	Parameters:	none
 *
 * 	Date:		April 24, 2004
 *
 * 	Author:		Steve Bell
 *
 * **************************************************************************************************************************************/

void xmet(void)
{
	INT32 scanReturn;
	char* xmet;
	if( firstTime == my_true ){
		xSize = numberOfMets;
		xFluxes = (char**) malloc((unsigned) xSize * sizeof(char*));
		xFluxTypes = (XFLXTYPE*) malloc((unsigned) xSize * sizeof(XFLXTYPE));
		if( (xFluxes == NULL) || (xFluxTypes == NULL) ){
			fprintf(stderr, "Cannot allocate exchange flux pointers\n");
			exit(4);
		}
		firstTime = my_false;
	}

	xmet = (char*) malloc((unsigned) 512 * sizeof(char));
	if( xmet == NULL ){
		fprintf(stderr, "Cannot allocate pointer");
		exit(1);
	}
	scanReturn = fscanf(fp, "%s", xmet);
	if( (scanReturn == 0) || (scanReturn == EOF) ){
		fprintf(stderr, "Error in input file, line %d\n", lineNumber);
		exit(2);
	}
	if( metListFind( xmet, metHeader, metCurrent ) ){
		xFluxes[numberOfXfluxes] = xmet;
		
	}
	else{
		fprintf(stderr, "Metabolite %s not found, line %d, \n", xmet, lineNumber);
		exit(2);
	}
	currentState = EX_TYPE;
} /*xmet()*/

/***************************************************************************************************************************************
 *
 * 	Function:	extype()
 *
 * 	Description:	Function enters the value of the exchange flux (input = -1, free = 0, output = 1) into the array
 *
 * 	Parameters:	none
 *
 * 	Date:		April 24, 2004
 *
 * 	Author:		Steve Bell
 *
 * **************************************************************************************************************************************/

void extype(void)
{
	INT32 scanReturn;
	UINT32 length;
	int inChar;
	char* extype = (char*) malloc((unsigned) 512 * sizeof(char));
	if( extype == NULL ){
		fprintf(stderr, "Cannot allocate pointer");
		exit(1);
	}
	scanReturn = fscanf(fp, "%s", extype);
	if( (scanReturn == 0) || (scanReturn == EOF) ){
		fprintf(stderr, "Error in input file, line %d\n", lineNumber);
		exit(2);
	}
	length = strlen(extype);
	if( extype[length - 1] == ';' ){
		extype[length - 1] = '\0';
	}
	extype = lowerCase(extype);
	if( strcmp(extype, "input") == 0 ){
		xFluxTypes[numberOfXfluxes] = INPUT;
	}
	else if( strcmp(extype, "free") == 0 ) {
		xFluxTypes[numberOfXfluxes] = FREE;
	}
	else if( strcmp(extype, "output") == 0 ){
		xFluxTypes[numberOfXfluxes] = OUTPUT;
	}
	else{
		fprintf(stderr, "Error in exchange flux type %s, line %d \n", extype, lineNumber);
		exit(2);
	}
	numberOfXfluxes++;
	if( numberOfXfluxes >= xSize ){
		xSize += numberOfMets;
		xFluxes = (char**) realloc((void*) xFluxes, (unsigned) xSize * sizeof(char*));
		xFluxTypes = (XFLXTYPE*) realloc((void*) xFluxTypes, (unsigned) xSize * sizeof(XFLXTYPE));
		if( (xFluxes == NULL) || (xFluxTypes == NULL) ){
			fprintf(stderr, "Cannot allocate exchange flux pointers\n");
			exit(4);
		}
	}
	
	inChar = fgetc(fp);
	switch(inChar){
		case EOF:
			endDummy = my_true;
			break;
		case '\n':
			currentState = NEWLINE;
			break;
		case 13: /*carriage return*/
			currentState = NEWLINE;
			break;
		default:
			ungetc(inChar, fp);
			currentState = NEWLINE;
	}
	free((void*) extype);
} /*extype()*/

/***************************************************************************************************************************************
 *
 * 	Function:	finish()
 *
 * 	Description:	Function frees all pointers not needed
 *
 * 	Parameters:	none
 *
 * 	Date:		April 24, 2004
 *
 * 	Author:		Steve Bell
 *
 * **************************************************************************************************************************************/

void finish(void)
{
	metListMakeEmpty(metHeader, metCurrent);
	free((void*) *metHeader);
	free((void*) metHeader);
	free((void*) metCurrent);
	rxnListMakeEmpty(rxnHeader, rxnCurrent);
	free((void*) *rxnHeader);
	free((void*) rxnHeader);
	free((void*) rxnCurrent);
} /*finish()*/


char* lowerCase(char* string2Lower)
{
	UINT32 stringLength = strlen(string2Lower);
	char* returnString = (char*) malloc( (unsigned) (stringLength + 1) * sizeof(char) );
	UINT32 character;
	for( character = 0; character < stringLength; character++ ){
		returnString[character] = tolower( (int) string2Lower[character] );
	}
	returnString[character] = '\0';
	return returnString;
}
		




